# Standard Library
import datetime
import sys
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, StringConstraints

# First Party
from resc_backend.db.model import DBfinding


class FindingBase(BaseModel):
    file_path: Annotated[str, StringConstraints(max_length=500)]
    line_number: Annotated[int, Field(gt=-1)]
    column_start: Annotated[int, Field(gt=-1)]
    column_end: Annotated[int, Field(gt=-1)]
    commit_id: Annotated[str, StringConstraints(max_length=120)]
    commit_message: str
    commit_timestamp: datetime.datetime
    author: Annotated[str, StringConstraints(max_length=200)]
    email: Annotated[str, StringConstraints(max_length=100)]
    event_sent_on: datetime.datetime | None = None
    rule_name: Annotated[str, StringConstraints(max_length=400)]


class FindingPatch(BaseModel):
    event_sent_on: datetime.datetime


class FindingCreate(FindingBase):
    repository_id: Annotated[int, Field(gt=0)]

    @classmethod
    def create_from_base_class(cls, base_object: FindingBase, repository_id: int):
        return cls(**(dict(base_object)), repository_id=repository_id)


class Finding(FindingBase):
    pass


class FindingRead(FindingCreate):
    id_: Annotated[int, Field(gt=0)]
    scan_ids: Annotated[list[Annotated[int, Field(gt=0)]], Field(min_length=None, max_length=sys.maxsize)] | None = None
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def create_from_db_entities(cls, db_finding: DBfinding, scan_ids: list[int]):
        return FindingRead(
            id_=db_finding.id_,
            file_path=db_finding.file_path,
            line_number=db_finding.line_number,
            column_start=db_finding.column_start,
            column_end=db_finding.column_end,
            commit_id=db_finding.commit_id,
            commit_message=db_finding.commit_message,
            commit_timestamp=db_finding.commit_timestamp,
            author=db_finding.author,
            email=db_finding.email,
            event_sent_on=db_finding.event_sent_on,
            rule_name=db_finding.rule_name,
            repository_id=db_finding.repository_id,
            scan_ids=scan_ids,
        )



# Third Party
from pydantic import BaseModel, ConfigDict


class AuditorMetric(BaseModel):
    auditor: str
    true_positive: int
    false_positive: int
    clarification_required: int
    not_accessible: int
    outdated: int
    not_analyzed: int
    total: int
    model_config = ConfigDict(from_attributes=True)



# Third Party
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field, StringConstraints


class DateCountModel(BaseModel):
    date_lable: Annotated[str, StringConstraints(max_length=100)]
    finding_count: Annotated[int, Field(gt=-1)] = 0
    model_config = ConfigDict(from_attributes=True)



# Standard Library

# Third Party
from typing import Annotated

from pydantic import BaseModel, ConfigDict, Field, StringConstraints


class RuleBase(BaseModel):
    rule_name: Annotated[str, StringConstraints(min_length=1, max_length=400)]
    description: Annotated[str, StringConstraints(max_length=4000)] | None = None
    comment: Annotated[str, StringConstraints(max_length=2000)] | None = None
    entropy: float | None = None
    secret_group: int | None = None
    regex: str | None = None
    path: str | None = None
    keywords: str | None = None


class RuleCreate(RuleBase):
    rule_pack: Annotated[str, StringConstraints(pattern=r"^(\d+\.)?(\d+\.)?(\*|\d+)$")]
    allow_list: Annotated[int, Field(gt=0)] | None = None

    @classmethod
    def create_from_base_class(cls, base_object: RuleBase, rule_pack: str, allow_list=int):
        return cls(**(dict(base_object)), rule_pack=rule_pack, allow_list=allow_list)


class Rule(RuleBase):
    pass


class RuleRead(RuleCreate):
    id_: Annotated[int, Field(gt=0)]
    model_config = ConfigDict(from_attributes=True)



# Third Party
from typing import Annotated

from pydantic import BaseModel, Field

# First Party
from resc_backend.resc_web_service.schema.auditor_metric import AuditorMetric


class PersonalAuditMetrics(BaseModel):
    today: Annotated[int, Field(gt=-1)] = 0
    current_week: Annotated[int, Field(gt=-1)] = 0
    last_week: Annotated[int, Field(gt=-1)] = 0
    current_month: Annotated[int, Field(gt=-1)] = 0
    current_year: Annotated[int, Field(gt=-1)] = 0
    forever: Annotated[int, Field(gt=-1)] = 0
    rank_current_week: Annotated[int, Field(gt=-1)] = 0
    forever_breakdown: AuditorMetric | None = None



# Standard Library
import logging
from datetime import UTC, datetime

# Third Party
from itertools import islice

from sqlalchemy import func, select, update
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import (
    DBaudit,
    DBfinding,
    DBrepository,
    DBscan,
    DBscanFinding,
    DBVcsInstance,
)
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import repository as repository_schema
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.scan_type import ScanType
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

logger = logging.getLogger(__name__)


def _get_max_base_scan(db_connection: Session) -> Query:
    subquery: Query = db_connection.query(DBscan.repository_id, func.max(DBscan.id_).label("latest_base_scan_id"))
    subquery = subquery.where(DBscan.scan_type == ScanType.BASE)
    subquery = subquery.where(DBscan.is_latest == True)  # noqa: E712
    subquery = subquery.group_by(DBscan.repository_id)
    return subquery.subquery()


def _only_if_has_untriaged_findings_condition(db_connection: Session) -> Query:
    has_untriaged_sub_query: Query = db_connection.query(DBfinding.repository_id)
    has_untriaged_sub_query = has_untriaged_sub_query.join(
        DBaudit,
        (DBaudit.finding_id == DBfinding.id_) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    has_untriaged_sub_query = has_untriaged_sub_query.where(
        (DBaudit.status == None) | (DBaudit.status == FindingStatus.NOT_ANALYZED)  # noqa: E711
    )  # noqa: E711
    return has_untriaged_sub_query.distinct()


def _repository_id_only_if_has_findings(db_connection: Session) -> Query:
    last_scan_sub_query = _get_max_base_scan(db_connection)

    sub_query = db_connection.query(DBrepository.id_)
    sub_query = sub_query.join(
        last_scan_sub_query,
        DBrepository.id_ == last_scan_sub_query.c.repository_id,
    )
    sub_query = sub_query.join(
        DBscan,
        DBrepository.id_ == DBscan.repository_id,
    )
    sub_query = sub_query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
    sub_query = sub_query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)
    return sub_query.distinct()


def _apply_filters(
    db_connection: Session,
    query: Query,
    include_deleted: bool,
    only_if_has_findings: bool,
    only_if_has_untriaged_findings: bool,
    vcs_providers: list[VCSProviders] | None,
) -> Query:
    if not include_deleted:
        query = query.filter(DBrepository.deleted_at == None)  # noqa: E711

    if only_if_has_findings:
        sub_query = _repository_id_only_if_has_findings(db_connection=db_connection)
        query = query.where(DBrepository.id_.in_(sub_query))

    if only_if_has_untriaged_findings:
        has_untriaged_sub_query = _only_if_has_untriaged_findings_condition(db_connection)
        query = query.where(DBrepository.id_.in_(has_untriaged_sub_query))

    if vcs_providers and vcs_providers is not None:
        query = query.where(DBVcsInstance.provider_type.in_(vcs_providers))

    return query


def get_repositories(
    db_connection: Session,
    vcs_providers: list[VCSProviders] | None = None,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    project_filter: str = "",
    repository_filter: str = "",
    only_if_has_findings: bool = False,
    include_deleted: bool = False,
    only_if_has_untriaged_findings: bool = False,
):
    """
        Retrieve repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name. Is used as a string contains filter
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: repositories
        list of DBrepository objects
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit

    # Get the latest scan for repository
    sub_query: Query = db_connection.query(DBscan.repository_id, func.max(DBscan.timestamp).label("max_timestamp"))
    sub_query = sub_query.where(DBscan.is_latest == True)  # noqa: E712
    sub_query = sub_query.group_by(DBscan.repository_id)
    sub_query = sub_query.subquery()

    query = db_connection.query(
        DBrepository.id_,
        DBrepository.project_key,
        DBrepository.repository_id,
        DBrepository.repository_name,
        DBrepository.repository_url,
        DBrepository.vcs_instance,
        DBrepository.deleted_at,
        DBVcsInstance.provider_type,
        func.coalesce(DBscan.id_, None).label("last_scan_id"),
        func.coalesce(DBscan.timestamp, None).label("last_scan_timestamp"),
    )

    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
    query = query.join(
        sub_query,
        DBrepository.id_ == sub_query.c.repository_id,
        isouter=True,
    )
    query = query.join(
        DBscan,
        (DBscan.repository_id == sub_query.c.repository_id) & (DBscan.timestamp == sub_query.c.max_timestamp),
        isouter=True,
    )

    query = _apply_filters(
        db_connection, query, include_deleted, only_if_has_findings, only_if_has_untriaged_findings, vcs_providers
    )

    if project_filter:
        query = query.where(DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    repositories = query.order_by(DBrepository.repository_name).offset(skip).limit(limit_val).all()

    return repositories


def get_repositories_count(
    db_connection: Session,
    vcs_providers: list[VCSProviders] = None,
    project_filter: str = "",
    repository_filter: str = "",
    only_if_has_findings: bool = False,
    include_deleted: bool = False,
    only_if_has_untriaged_findings: bool = False,
) -> int:
    """
        Retrieve count of repository records optionally filtered
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional [string] filtering the VCS provider
    :param project_filter:
        optional, filter on project name
    :param repository_filter:
        optional, filter on repository name
    :param only_if_has_findings:
        optional, filter on repositories with findings
    :return: total_count
        count of repositories
    """
    query = db_connection.query(func.count(DBrepository.id_))
    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)

    query = _apply_filters(
        db_connection, query, include_deleted, only_if_has_findings, only_if_has_untriaged_findings, vcs_providers
    )

    if project_filter:
        query = query.where(DBrepository.project_key == project_filter)

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    total_count = query.scalar()
    return total_count


def get_repository(db_connection: Session, repository_id: int):
    repository = db_connection.query(DBrepository).where(DBrepository.id_ == repository_id).first()
    return repository


def update_repository(
    db_connection: Session,
    repository_id: int,
    repository: repository_schema.RepositoryCreate,
):
    db_repository = db_connection.query(DBrepository).filter_by(id_=repository_id).first()

    db_repository.repository_name = repository.repository_name
    db_repository.repository_url = str(repository.repository_url)
    db_repository.vcs_instance = repository.vcs_instance
    db_repository.deleted_at = repository.deleted_at

    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository(db_connection: Session, repository: repository_schema.RepositoryCreate):
    db_repository = DBrepository(
        project_key=repository.project_key,
        repository_id=repository.repository_id,
        repository_name=repository.repository_name,
        repository_url=str(repository.repository_url),
        vcs_instance=repository.vcs_instance,
        deleted_at=None,
    )
    db_connection.add(db_repository)
    db_connection.commit()
    db_connection.refresh(db_repository)
    return db_repository


def create_repository_if_not_exists(db_connection: Session, repository: repository_schema.RepositoryCreate):
    # Query the database to see if the repository object exists based on the unique constraint parameters
    query: Query = db_connection.query(DBrepository)
    query = query.where(DBrepository.project_key == repository.project_key)
    query = query.where(DBrepository.repository_id == repository.repository_id)
    query = query.where(DBrepository.vcs_instance == repository.vcs_instance)
    db_select_repository = query.first()

    if db_select_repository is not None:
        return db_select_repository

    # Create non-existing repository object
    return create_repository(db_connection, repository)


def get_distinct_projects(
    db_connection: Session,
    vcs_providers: list[VCSProviders] = None,
    repository_filter: str = "",
    only_if_has_findings: bool = False,
    include_deleted: bool = False,
    only_if_has_untriaged_findings: bool = False,
):
    """
        Retrieve all unique project names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param repository_filter:
        optional, filter on repository name. Is used as a string contains filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_projects
        The output will contain a list of unique projects
    """
    query = db_connection.query(DBrepository.project_key)
    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)

    query = _apply_filters(
        db_connection, query, include_deleted, only_if_has_findings, only_if_has_untriaged_findings, vcs_providers
    )

    if repository_filter:
        query = query.where(DBrepository.repository_name == repository_filter)

    query = query.distinct()
    query = query.order_by(DBrepository.project_key)
    distinct_projects = query.all()
    return distinct_projects


def get_distinct_repositories(
    db_connection: Session,
    vcs_providers: list[VCSProviders] = None,
    project_name: str = "",
    only_if_has_findings: bool = False,
    include_deleted: bool = False,
    only_if_has_untriaged_findings: bool = False,
):
    """
        Retrieve all unique repository names
    :param db_connection:
        Session of the database connection
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param project_name:
        optional, filter on project name. Is used as a full string match filter
    :param only_if_has_findings:
        optional, filter on repositories that have findings
    :return: distinct_repositories
        The output will contain a list of unique repositories
    """
    query = db_connection.query(DBrepository.repository_name)
    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)

    query = _apply_filters(
        db_connection, query, include_deleted, only_if_has_findings, only_if_has_untriaged_findings, vcs_providers
    )

    if project_name:
        query = query.where(DBrepository.project_key == project_name)

    query = query.distinct()
    query = query.order_by(DBrepository.repository_name)
    distinct_repositories = query.all()
    return distinct_repositories


def get_findings_metadata_by_repository_id(db_connection: Session, repository_ids: list[int]):
    """
        Retrieves the finding metadata for a repository id from the database with most recent scan information
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        ids of the repository for which findings metadata to be retrieved
    :return: findings_metadata
        findings_metadata containing the count for each status
    """
    query: Query = db_connection.query(DBrepository.id_, DBaudit.status, func.count(DBscanFinding.finding_id))

    last_scan_sub_query = _get_max_base_scan(db_connection)
    query = query.join(
        last_scan_sub_query,
        DBrepository.id_ == last_scan_sub_query.c.repository_id,
    )
    query = query.join(
        DBscan,
        DBrepository.id_ == DBscan.repository_id,
    )
    query = query.where(DBscan.id_ >= last_scan_sub_query.c.latest_base_scan_id)
    query = query.join(DBscanFinding, DBscan.id_ == DBscanFinding.scan_id)
    query = query.join(
        DBaudit,
        (DBaudit.finding_id == DBscanFinding.finding_id) & (DBaudit.is_latest == True),  # noqa: E712
        isouter=True,
    )
    query = query.where(DBrepository.id_.in_(repository_ids))
    query = query.group_by(
        DBrepository.id_,
        DBaudit.status,
    )
    status_counts = query.all()
    repo_count_dict = {}
    for repository_id in repository_ids:
        repo_count_dict[repository_id] = FindingStatus.init_statistics()

    for status_count in status_counts:
        repository_id: str = status_count[0]
        finding_status: str | None = status_count[1]
        count: int = status_count[2]

        repo_count_dict[repository_id]["total_findings_count"] += count
        if finding_status is None:
            repo_count_dict[repository_id][FindingStatus.NOT_ANALYZED.value.lower()] += count
        else:
            repo_count_dict[repository_id][finding_status.lower()] += count

    return repo_count_dict


def delete_repository(db_connection: Session, repository_id: int, delete_related: bool = False):
    """
        Delete a repository object
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding_by_repository_id(db_connection, repository_id=repository_id)
        finding_crud.delete_findings_by_repository_id(db_connection, repository_id=repository_id)
        scan_crud.delete_scans_by_repository_id(db_connection, repository_id=repository_id)
    db_connection.query(DBrepository).where(DBrepository.id_ == repository_id).delete(synchronize_session=False)
    db_connection.commit()


def delete_repositories_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete repositories for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(DBrepository).where(
        DBrepository.vcs_instance == DBVcsInstance.id_,
        DBVcsInstance.id_ == vcs_instance_id,
    ).delete(synchronize_session=False)
    db_connection.commit()


def soft_delete_repository(db_connection: Session, repository_ids: list[int]):
    """
        Soft delete a repository object
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        list of id of the repository to be deleted
    """
    iterator = iter(repository_ids)
    while chunk := list(islice(iterator, 1000)):
        db_connection.execute(
            update(DBrepository).where(DBrepository.id_.in_(chunk)).values(deleted_at=datetime.now(UTC))
        )
    db_connection.commit()


def undelete_repository(db_connection: Session, repository_ids: list[int]):
    """
        Undelete a repository object
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        list of id of the repository to be undeleted
    """
    iterator = iter(repository_ids)
    while chunk := list(islice(iterator, 1000)):
        db_connection.execute(update(DBrepository).where(DBrepository.id_.in_(chunk)).values(deleted_at=None))
    db_connection.commit()


def get_active_repository_ids_by_project_and_vcs_instance(
    db_connection: Session, project_key: str, vcs_instance_name: str
) -> list[str]:
    logger.debug(f"Fetching active repository ids for project {project_key} and vcs instance {vcs_instance_name}")
    query = select(DBrepository.repository_id)
    query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
    query = query.where(DBrepository.project_key == project_key)
    query = query.where(DBVcsInstance.name == vcs_instance_name)
    query = query.where(DBrepository.deleted_at.is_(None))
    return db_connection.execute(query).scalars().all()


def fetch_id_from_undeleted_repository_string_id(
    db_connection: Session, vcs_instance_name: str, repository_ids: list[str]
) -> list[int]:
    """
        Fetch the id of the undeleted repository from the string id
    :param db_connection:
        Session of the database connection
    :param repository_ids:
        list of id of the repository
    """
    iterator = iter(repository_ids)
    repository_ids = []
    while chunk := list(islice(iterator, 1000)):
        query = select(DBrepository.id_)
        query = query.join(DBVcsInstance, DBVcsInstance.id_ == DBrepository.vcs_instance)
        query = query.where(DBrepository.repository_id.in_(chunk))
        query = query.where(DBVcsInstance.name == vcs_instance_name)
        query = query.where(DBrepository.deleted_at.is_(None))
        repository_ids.extend(db_connection.execute(query).scalars().all())
    return repository_ids



# Standard Library
import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, StringConstraints

# First Party
from resc_backend.constants import MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.resc_web_service.schema.finding_status import FindingStatus


class AuditMultiple(BaseModel):
    finding_ids: Annotated[
        list[Annotated[int, Field(gt=0)]], Field(min_length=1, max_length=MAX_RECORDS_PER_PAGE_LIMIT)
    ]
    status: FindingStatus
    comment: Annotated[str, StringConstraints(max_length=255)]


class AuditRead(BaseModel):
    id_: Annotated[int, Field(gt=0)]
    status: FindingStatus
    auditor: Annotated[str, StringConstraints(max_length=250)]
    comment: Annotated[str, StringConstraints(max_length=255)] | None = None
    timestamp: datetime.datetime
    model_config = ConfigDict(from_attributes=True)



# Standard Library
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, StringConstraints, ValidationInfo, field_validator

# First Party
from resc_backend.constants import AZURE_DEVOPS
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class VCSInstanceBase(BaseModel):
    name: Annotated[str, StringConstraints(max_length=200)]
    provider_type: VCSProviders
    hostname: Annotated[str, StringConstraints(max_length=200)]
    port: Annotated[int, Field(gt=-0, lt=65536)]
    scheme: Annotated[str, StringConstraints(max_length=20)]
    exceptions: Annotated[list[str], Field(min_length=None, max_length=500)] | None = None
    scope: Annotated[list[str], Field(min_length=None, max_length=500)] | None = None
    organization: Annotated[str, StringConstraints(max_length=200)] | None = None

    @field_validator("scheme", mode="before")
    @classmethod
    def check_scheme(cls, value):
        allowed_schemes = ["http", "https"]
        if value not in allowed_schemes:
            raise ValueError(f"The scheme '{value}' must be one of the following {', '.join(allowed_schemes)}")
        return value

    @field_validator("organization", mode="before")
    @classmethod
    def check_organization(cls, value, values: ValidationInfo):
        if not value and values.data["provider_type"] == AZURE_DEVOPS:
            raise ValueError("The organization field needs to be specified for Azure devops vcs instances")
        return value

    @field_validator("scope", mode="before")
    @classmethod
    def check_scope_and_exceptions(cls, value, values: ValidationInfo):
        if value and values.data["exceptions"]:
            raise ValueError(
                "You cannot specify both the scope and exceptions to the scan, only one setting is supported."
            )
        return value


class VCSInstanceCreate(VCSInstanceBase):
    pass


class VCSInstanceRead(VCSInstanceBase):
    id_: Annotated[int, Field(gt=0)]

    @classmethod
    def create_from_db_vcs_instance(cls, db_vcs_instance):
        exceptions = []
        scope = []
        if db_vcs_instance.exceptions:
            exceptions = db_vcs_instance.exceptions.split(",")
        if db_vcs_instance.scope:
            scope = db_vcs_instance.scope.split(",")

        vcs_instance_read = cls(
            id_=db_vcs_instance.id_,
            name=db_vcs_instance.name,
            provider_type=db_vcs_instance.provider_type,
            hostname=db_vcs_instance.hostname,
            port=db_vcs_instance.port,
            scheme=db_vcs_instance.scheme,
            exceptions=exceptions,
            scope=scope,
            organization=db_vcs_instance.organization,
        )

        return vcs_instance_read

    model_config = ConfigDict(from_attributes=True)



# Standard Library
import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, StringConstraints

RULE_PACK_VERSION_REGEX = r"^\d+(?:\.\d+){2}$"


class RulePackBase(BaseModel):
    version: Annotated[str, StringConstraints(pattern=RULE_PACK_VERSION_REGEX)]
    active: bool = False
    global_allow_list: Annotated[int, Field(gt=0)] | None = None
    outdated: bool = False


class RulePackCreate(RulePackBase):
    version: Annotated[str, StringConstraints(pattern=RULE_PACK_VERSION_REGEX)]
    global_allow_list: Annotated[int, Field(gt=0)] | None = None

    @classmethod
    def create_from_base_class(cls, base_object: RulePackBase, global_allow_list: int):
        return cls(**(dict(base_object)), global_allow_list=global_allow_list)


class RulePack(RulePackBase):
    pass


class RulePackRead(RulePackBase):
    version: Annotated[str, StringConstraints(pattern=RULE_PACK_VERSION_REGEX)]
    created: datetime.datetime
    model_config = ConfigDict(from_attributes=True)


class RulePackVersion(BaseModel):
    version: Annotated[str, StringConstraints(pattern=RULE_PACK_VERSION_REGEX)]



# Standard Library
from typing import Annotated, Generic, TypeVar

# Third Party
from pydantic import BaseModel, ConfigDict, Field

Model = TypeVar("Model", bound=BaseModel)


class FindingCountModel(BaseModel, Generic[Model]):
    """
        Generic encapsulation class for findings count end points to standardize output of the API
        example creation, FindingCountModel[FindingRead](data=db_findings, true_positive=true_positive,
        false_positive=false_positive, not_analyzed=not_analyzed, not_accessible=not_accessible,
        clarification_required=clarification_required, outdated=outdated,
        total_findings_count=total_findings_count)
    :param Generic[Model]:
        Type of the object in the data list
    """

    data: Model | None
    true_positive: Annotated[int, Field(gt=-1)]
    false_positive: Annotated[int, Field(gt=-1)]
    not_analyzed: Annotated[int, Field(gt=-1)]
    not_accessible: Annotated[int, Field(gt=-1)]
    clarification_required: Annotated[int, Field(gt=-1)]
    outdated: Annotated[int, Field(gt=-1)]
    total_findings_count: Annotated[int, Field(gt=-1)]
    model_config = ConfigDict(from_attributes=True)



# Standard Library
from typing import Annotated, Generic, TypeVar

# Third Party
# pylint: disable=no-name-in-module
from pydantic import BaseModel, ConfigDict, Field

# First Party
from resc_backend.constants import MAX_RECORDS_PER_PAGE_LIMIT

Model = TypeVar("Model", bound=BaseModel)


class PaginationModel(BaseModel, Generic[Model]):
    """
        Generic encapsulation class for paginated endpoints to standardize output of the API
        example creation, PaginationModel[FindingRead](data=db_findings, total=total, limit=limit, skip=skip)
    :param Generic[Model]:
        Type of the object in the data list
    """

    # data: List[Model]
    data: Annotated[list[Model], Field(min_length=None, max_length=MAX_RECORDS_PER_PAGE_LIMIT)]
    total: Annotated[int, Field(gt=-1)]
    limit: Annotated[int, Field(gt=-1)]
    skip: Annotated[int, Field(gt=-1)]
    model_config = ConfigDict(from_attributes=True)



# Third Party
import tomlkit
from tomlkit import aot, comment, document, nl, table
from tomlkit.items import AoT, Array, String, StringType, Table
from tomlkit.toml_document import TOMLDocument

# First Party
from resc_backend.constants import (
    TEMP_RULE_FILE,
    TOML_CUSTOM_DELIMITER,
)


def create_toml_rule_file(parsed_toml_dictionary: dict):
    """
        Create a TOML file from a dictionary
    :param parsed_toml_dictionary:
        TOML dictionary
    :return: toml_file
        Returns toml file
    """
    doc = create_toml_document(parsed_toml_dictionary)
    return write_toml_document(doc, TEMP_RULE_FILE)


def create_toml_document(parsed_toml_dictionary: dict) -> TOMLDocument:
    """
        Create a TOML document from a dictionary
    :param parsed_toml_dictionary:
        TOML dictionary
    :return: TOMLDocument
        Returns toml document in memory
    """
    doc = document()
    doc.add(comment("This is a gitleaks configuration file."))
    doc.add(comment("Rules and allowlists are defined within this file."))
    doc.add(comment("Rules instruct gitleaks on what should be considered a secret."))
    doc.add(comment("Allowlists instruct gitleaks on what is allowed, i.e. not a secret."))
    doc.add(nl())

    if "title" in parsed_toml_dictionary:
        doc.add("title", parsed_toml_dictionary["title"])
    doc.add(nl())
    if "version" in parsed_toml_dictionary:
        doc.add("version", parsed_toml_dictionary["version"])
    doc.add(nl())

    # Global allow list table
    global_allow_list_table = create_allow_list_toml_table(input_dictionary=parsed_toml_dictionary, key="allowlist")
    if global_allow_list_table:
        doc.add("allowlist", global_allow_list_table)
        doc.add(nl())

    # Rules table
    rule_array_table = create_rule_array_toml_table(rule_dictionary=parsed_toml_dictionary)
    doc.add("rules", rule_array_table)
    return doc


def write_toml_document(doc: TOMLDocument, destination: str = TEMP_RULE_FILE):
    """
        Write a TOML document to a TOML file
    :param doc:
        TOMLDocument document
    :param destination:
        file name of the returned document
    :return: toml_file
        Returns toml file
    """
    toml_string = tomlkit.dumps(doc)
    with open(destination, "w", encoding="utf-8") as toml_file:
        toml_file.write(toml_string)
        toml_file.close()
    return toml_file


def get_multiline_array_for_toml_file(input_dictionary: dict, key: str, string_type: str, delimiter: str) -> Array:
    """
        Create multiline toml array for the input dictionary value
    :param input_dictionary:
        Input dictionary
    :param key:
        key of Input dictionary
    :param string_type:
        Multi Line Literal or Single Line Basic
    :param delimiter:
        TOML_CUSTOM_DELIMITER or ","
    :return: tomlkit.array
        The output will return a toml array
    """
    multiline_array = tomlkit.array()
    array_from_db = input_dictionary[key].split(delimiter)
    for value_str in array_from_db:
        multiline_array.append(String.from_raw(String.from_raw(value_str), string_type))
        multiline_array.multiline(True)
    return multiline_array


def create_allow_list_toml_table(input_dictionary: dict, key: str) -> Table:
    """
        Create a TOML table for rule allow list
    :param input_dictionary:
        AllowList dictionary
     :param key:
        Key in AllowList dictionary
    :return: table
        Returns allow list TOML table
    """
    allow_list_table = None
    if key in input_dictionary:
        allow_list_table = table()
        allow_list_dict = input_dictionary[key]
        if "description" in allow_list_dict:
            allow_list_table.add("description", allow_list_dict["description"])
        if "paths" in allow_list_dict:
            multiline_path_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="paths",
                string_type=StringType.MLL,
                delimiter=TOML_CUSTOM_DELIMITER,
            )
            allow_list_table.add("paths", multiline_path_array)
        if "regexes" in allow_list_dict:
            multiline_regex_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="regexes",
                string_type=StringType.MLL,
                delimiter=TOML_CUSTOM_DELIMITER,
            )
            allow_list_table.add("regexes", multiline_regex_array)
        if "commits" in allow_list_dict:
            multiline_commit_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="commits",
                string_type=StringType.SLB,
                delimiter=",",
            )
            allow_list_table.add("commits", multiline_commit_array)
        if "stop_words" in allow_list_dict:
            multiline_stopword_array = get_multiline_array_for_toml_file(
                input_dictionary=allow_list_dict,
                key="stop_words",
                string_type=StringType.SLB,
                delimiter=",",
            )
            allow_list_table.add("stopwords", multiline_stopword_array)
    return allow_list_table


def create_rule_array_toml_table(rule_dictionary: dict) -> AoT:
    """
       Create an array of table for rule list
    :param rule_dictionary:
        Rule dictionary
    :return: table
        Return an array of table
    """

    # Rule Table
    rule_array_table = aot()
    if "rules" in rule_dictionary:
        for rule_dict in rule_dictionary["rules"]:
            rule_table = table()
            if "id" in rule_dict:
                rule_table.add("id", rule_dict["id"])
            if "description" in rule_dict:
                rule_table.add("description", rule_dict["description"])
            if "entropy" in rule_dict:
                rule_table.add("entropy", rule_dict["entropy"])
            if "secret_group" in rule_dict:
                rule_table.add("secretGroup", rule_dict["secret_group"])
            if "regex" in rule_dict:
                rule_table.add("regex", String.from_raw(rule_dict["regex"], StringType.MLL))
            if "path" in rule_dict:
                rule_table.add("path", String.from_raw(rule_dict["path"], StringType.MLL))
            if "tags" in rule_dict:
                multiline_tag_array = get_multiline_array_for_toml_file(
                    input_dictionary=rule_dict,
                    key="tags",
                    string_type=StringType.SLB,
                    delimiter=",",
                )
                rule_table.add("tags", multiline_tag_array)
            if "keywords" in rule_dict:
                multiline_keyword_array = get_multiline_array_for_toml_file(
                    input_dictionary=rule_dict,
                    key="keywords",
                    string_type=StringType.SLB,
                    delimiter=",",
                )
                rule_table.add("keywords", multiline_keyword_array)
            if "comment" in rule_dict:
                rule_table.add("comment", rule_dict["comment"])

            # Rule Allow List Table
            if "allow_list" in rule_dict:
                allow_list_table = create_allow_list_toml_table(input_dictionary=rule_dict, key="allow_list")
                rule_table.append("allowlist", allow_list_table)

            rule_array_table.append(rule_table)
    return rule_array_table



# Standard Library
import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, HttpUrl, StringConstraints

# First Party
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


class RepositoryEnrichedBase(BaseModel):
    project_key: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_id: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_name: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    repository_url: HttpUrl
    vcs_provider: VCSProviders
    last_scan_id: Annotated[int, Field(gt=0)] | None = None
    last_scan_timestamp: datetime.datetime | None = None
    true_positive: Annotated[int, Field(gt=-1)]
    false_positive: Annotated[int, Field(gt=-1)]
    not_analyzed: Annotated[int, Field(gt=-1)]
    not_accessible: Annotated[int, Field(gt=-1)]
    clarification_required: Annotated[int, Field(gt=-1)]
    outdated: Annotated[int, Field(gt=-1)]
    total_findings_count: Annotated[int, Field(gt=-1)]
    deleted_at: datetime.datetime | None = None


class RepositoryEnriched(RepositoryEnrichedBase):
    pass


class RepositoryEnrichedRead(RepositoryEnriched):
    id_: Annotated[int, Field(gt=0)]
    model_config = ConfigDict(from_attributes=True)



# Standard Library
import logging.config
import sysconfig
from os import path

logger = logging.getLogger(__name__)


def get_logging_settings_path():
    if path.isfile(sysconfig.get_path("purelib") + "/resc"):
        base_dir = sysconfig.get_path("purelib") + "/resc"
    else:
        base_dir = path.dirname(__file__)

    return base_dir + "/static/logging.ini"


def initialise_logs(log_file_path: str, debug: bool = True):
    logging_ini_file = get_logging_settings_path()
    logging.config.fileConfig(
        logging_ini_file,
        defaults={"log_file_path": log_file_path},
        disable_existing_loggers=False,
    )
    logger_config = logging.getLogger("root")
    if debug:
        logger_config.setLevel(logging.DEBUG)
    else:
        logger_config.setLevel(logging.INFO)
    return logger_config



# Standard Library
import datetime
from typing import Annotated

# Third Party
from pydantic import BaseModel, ConfigDict, Field, StringConstraints

# First Party
from resc_backend.resc_web_service.schema.scan_type import ScanType


class ScanBase(BaseModel):
    scan_type: ScanType = ScanType.BASE
    last_scanned_commit: Annotated[str, StringConstraints(min_length=1, max_length=100)]
    timestamp: datetime.datetime
    increment_number: Annotated[int, Field(gt=-1)] = 0
    rule_pack: Annotated[str, StringConstraints(max_length=100)]


class ScanCreate(ScanBase):
    repository_id: Annotated[int, Field(gt=0)]

    @classmethod
    def create_from_base_class(cls, base_object: ScanBase, repository_id: int):
        return cls(**(dict(base_object)), repository_id=repository_id)


class Scan(ScanBase):
    pass


class ScanRead(ScanCreate):
    id_: Annotated[int, Field(gt=0)]
    model_config = ConfigDict(from_attributes=True)
